"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getCorrespondingJavascriptType2 = getCorrespondingJavascriptType2;
exports.getCorrespondingJavascriptType = getCorrespondingJavascriptType;
const node_opcua_variant_1 = require("node-opcua-variant");
const cache_1 = require("./cache");
const utils_1 = require("./utils");
async function getCorrespondingJavascriptType2(session, nodeId, dataTypeNodeId, cache, importCollect) {
    const { dataType, jtype, dataTypeCombination, type } = await getCorrespondingJavascriptType(session, dataTypeNodeId, cache, importCollect);
    const valueRank = await (0, utils_1.getValueRank)(session, nodeId);
    let jtype2 = "";
    if (valueRank >= 0) {
        jtype2 = jtype + "[]";
    }
    else if (valueRank === -1) {
        jtype2 = jtype;
    }
    else if (valueRank === -2) {
        //  The value can be a scalar or an array with any number of dimensions.
        jtype2 = `(${jtype} | ${jtype}[])`;
    }
    else if (valueRank === -3) {
        // ScalarOrOneDimension (-3):  The value can be a scalar or a one dimensional array.
        jtype2 = `(${jtype} | ${jtype}[])`;
    }
    else {
        throw new Error("Invalid valueRank " + valueRank);
    }
    return { dataType: dataType, jtype: jtype2, dataTypeCombination, type };
}
// eslint-disable-next-line complexity
async function getCorrespondingJavascriptType(session, dataTypeNodeId, cache, importCollect) {
    const info = await (0, utils_1._convertNodeIdToDataTypeAsync)(session, dataTypeNodeId);
    const { type } = info;
    if (type == "enum" && info.enumerationType) {
        // we have a enmeration name here
        const jtypeImport = await (0, cache_1.referenceEnumeration)(session, dataTypeNodeId);
        const jtype = jtypeImport.name;
        importCollect && importCollect(jtypeImport);
        return { dataType: info.dataType, jtype, type };
    }
    if (type == "basic" && info.dataType === node_opcua_variant_1.DataType.ExtensionObject) {
        const jtypeImport = await (0, cache_1.referenceExtensionObject)(session, dataTypeNodeId);
        const jtype = jtypeImport.name;
        importCollect && importCollect(jtypeImport);
        return { dataType: info.dataType, jtype, type };
    }
    const referenceBasicType = (name) => {
        const t = { name, namespace: -1, module: "BasicType" };
        importCollect && importCollect(t);
        cache.ensureImported(t);
        return t.name;
    };
    if (type == "genericNumber") {
        const { dataTypeCombination } = info;
        return { dataType: node_opcua_variant_1.DataType.Variant, jtype: "number", type, dataTypeCombination };
    }
    else {
        const { dataType } = info;
        switch (dataType) {
            case node_opcua_variant_1.DataType.Null:
                return { dataType: node_opcua_variant_1.DataType.Variant, jtype: referenceBasicType("VariantOptions"), type };
            case node_opcua_variant_1.DataType.Boolean:
                return { dataType, jtype: "boolean", type };
            case node_opcua_variant_1.DataType.Byte:
                return { dataType, jtype: referenceBasicType("Byte"), type };
            case node_opcua_variant_1.DataType.ByteString:
                return { dataType, jtype: "Buffer", type };
            case node_opcua_variant_1.DataType.DataValue:
                return { dataType, jtype: referenceBasicType("DataValue"), type };
            case node_opcua_variant_1.DataType.DateTime:
                return { dataType, jtype: "Date", type };
            case node_opcua_variant_1.DataType.DiagnosticInfo:
                return { dataType, jtype: referenceBasicType("DiagnosticInfo"), type };
            case node_opcua_variant_1.DataType.Double:
                return { dataType, jtype: "number", type };
            case node_opcua_variant_1.DataType.Float:
                return { dataType, jtype: "number", type };
            case node_opcua_variant_1.DataType.Guid:
                return { dataType, jtype: referenceBasicType("Guid"), type };
            case node_opcua_variant_1.DataType.Int16:
                return { dataType, jtype: referenceBasicType("Int16"), type };
            case node_opcua_variant_1.DataType.Int32:
                return { dataType, jtype: referenceBasicType("Int32"), type };
            case node_opcua_variant_1.DataType.UInt16:
                return { dataType, jtype: referenceBasicType("UInt16"), type };
            case node_opcua_variant_1.DataType.UInt32:
                return { dataType, jtype: referenceBasicType("UInt32"), type };
            case node_opcua_variant_1.DataType.UInt64:
                return { dataType, jtype: referenceBasicType("UInt64"), type };
            case node_opcua_variant_1.DataType.Int64:
                return { dataType, jtype: referenceBasicType("Int64"), type };
            case node_opcua_variant_1.DataType.LocalizedText:
                return { dataType, jtype: referenceBasicType("LocalizedText"), type };
            case node_opcua_variant_1.DataType.NodeId:
                return { dataType, jtype: referenceBasicType("NodeId"), type };
            case node_opcua_variant_1.DataType.ExpandedNodeId:
                return { dataType, jtype: referenceBasicType("ExpandedNodeId"), type };
            case node_opcua_variant_1.DataType.QualifiedName:
                return { dataType, jtype: referenceBasicType("QualifiedName"), type };
            case node_opcua_variant_1.DataType.SByte:
                return { dataType, jtype: referenceBasicType("SByte"), type };
            case node_opcua_variant_1.DataType.StatusCode:
                return { dataType, jtype: referenceBasicType("StatusCode"), type };
            case node_opcua_variant_1.DataType.String:
                return { dataType, jtype: referenceBasicType("UAString"), type };
            case node_opcua_variant_1.DataType.Variant:
                return { dataType, jtype: referenceBasicType("Variant"), type };
            case node_opcua_variant_1.DataType.XmlElement:
                return { dataType, jtype: referenceBasicType("String"), type };
            default:
                throw new Error("Unsupported " + dataType + " " + node_opcua_variant_1.DataType[dataType]);
        }
    }
}
//# sourceMappingURL=get_corresponding_data_type.js.map