"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDefinition = getDefinition;
exports.getBrowseName = getBrowseName;
exports.getDataTypeNodeId = getDataTypeNodeId;
exports.getIsAbstract = getIsAbstract;
exports.getDescription = getDescription;
exports.getNodeClass = getNodeClass;
exports.getChildren = getChildren;
exports.getFolderElements = getFolderElements;
exports.getModellingRule = getModellingRule;
exports.isExtensionObject = isExtensionObject;
exports.isEnumeration = isEnumeration;
exports.extractBasicDataType = extractBasicDataType;
exports.getSubtypeNodeIdIfAny = getSubtypeNodeIdIfAny;
exports.getSubtypeNodeId = getSubtypeNodeId;
exports.getTypeDefOrBaseType = getTypeDefOrBaseType;
exports.getTypeDefinition = getTypeDefinition;
exports._convertNodeIdToDataTypeAsync = _convertNodeIdToDataTypeAsync;
exports.getChildrenOrFolderElements = getChildrenOrFolderElements;
exports.getValueRank = getValueRank;
const node_opcua_data_model_1 = require("node-opcua-data-model");
const node_opcua_nodeid_1 = require("node-opcua-nodeid");
const node_opcua_constants_1 = require("node-opcua-constants");
const node_opcua_types_1 = require("node-opcua-types");
const node_opcua_variant_1 = require("node-opcua-variant");
const node_opcua_debug_1 = require("node-opcua-debug");
const debugLog = (0, node_opcua_debug_1.make_debugLog)(__filename);
async function getDefinition(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.DataTypeDefinition });
    return dataValue.value.value || null;
}
async function getBrowseName(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.BrowseName });
    return dataValue.value.value;
}
async function getDataTypeNodeId(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.DataType });
    return dataValue.value.value || null;
}
async function getIsAbstract(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.IsAbstract });
    return dataValue.value.value;
}
async function getDescription(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.Description });
    return dataValue.value.value;
}
async function getNodeClass(session, nodeId) {
    const dataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.NodeClass });
    return dataValue.value.value;
}
async function getChildren(session, nodeId) {
    const browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Forward,
        includeSubtypes: true,
        nodeClassMask: node_opcua_data_model_1.NodeClassMask.Method | node_opcua_data_model_1.NodeClassMask.Object | node_opcua_data_model_1.NodeClass.Variable,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasChild"),
        resultMask: 0xffff
    });
    return browseResult.references || [];
}
async function getFolderElements(session, nodeId) {
    const browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Forward,
        includeSubtypes: true,
        nodeClassMask: node_opcua_data_model_1.NodeClassMask.Method | node_opcua_data_model_1.NodeClassMask.Object | node_opcua_data_model_1.NodeClass.Variable,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("Organizes"),
        resultMask: 0xffff
    });
    return browseResult.references || [];
}
async function getModellingRule(session, nodeId) {
    const browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Forward,
        includeSubtypes: true,
        nodeClassMask: node_opcua_data_model_1.NodeClassMask.Method | node_opcua_data_model_1.NodeClassMask.Object | node_opcua_data_model_1.NodeClass.Variable,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasModellingRule"),
        resultMask: 0xffff
    });
    if (!browseResult.references || browseResult.references.length === 0) {
        return null;
    }
    return browseResult.references[0].browseName.name;
}
async function isExtensionObject(session, nodeId) {
    const n = nodeId;
    if (n.namespace === 0 && n.identifierType === node_opcua_nodeid_1.NodeIdType.NUMERIC && n.value === node_opcua_constants_1.DataTypeIds.Structure) {
        return true;
    }
    if (n.namespace === 0 && n.identifierType === node_opcua_nodeid_1.NodeIdType.NUMERIC && n.value <= node_opcua_constants_1.DataTypeIds.DiagnosticInfo) {
        return false;
    }
    const r = await getSubtypeNodeIdIfAny(session, nodeId);
    return await isExtensionObject(session, r.nodeId);
}
async function isEnumeration(session, nodeId) {
    const n = nodeId;
    if (n.namespace === 0 && n.identifierType === node_opcua_nodeid_1.NodeIdType.NUMERIC && n.value === node_opcua_constants_1.DataTypeIds.Enumeration) {
        return true;
    }
    if (n.namespace === 0 && n.identifierType === node_opcua_nodeid_1.NodeIdType.NUMERIC && n.value <= node_opcua_constants_1.DataTypeIds.DiagnosticInfo) {
        return false;
    }
    const r = await getSubtypeNodeIdIfAny(session, nodeId);
    return await isEnumeration(session, r.nodeId);
}
async function extractBasicDataType(session, dataTypeNodeId) {
    const n = dataTypeNodeId;
    if (n.namespace === 0 && n.identifierType === node_opcua_nodeid_1.NodeIdType.NUMERIC && n.value <= node_opcua_constants_1.DataTypeIds.DiagnosticInfo) {
        return dataTypeNodeId.value;
    }
    const r = await getSubtypeNodeIdIfAny(session, dataTypeNodeId);
    return await extractBasicDataType(session, r.nodeId);
}
async function getSubtypeNodeIdIfAny(session, nodeId) {
    if (nodeId.isEmpty()) {
        return null;
    }
    const browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Inverse,
        includeSubtypes: true,
        nodeClassMask: 0, // NodeClassMask.ObjectType| NodeClass.VariableType,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasSubtype"),
        resultMask: 0xffff
    });
    if (!browseResult.references || browseResult.references.length === 0) {
        return null;
    }
    return browseResult.references[0];
}
async function getSubtypeNodeId(session, nodeId) {
    const r = await getSubtypeNodeIdIfAny(session, nodeId);
    if (!r) {
        throw new Error("No Subtype");
    }
    return r;
}
async function getTypeDefOrBaseType(session, nodeId) {
    let browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Forward,
        includeSubtypes: true,
        nodeClassMask: 0,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasTypeDefinition"),
        resultMask: 0xffff
    });
    if (!browseResult.references || browseResult.references.length === 0) {
        browseResult = await session.browse({
            browseDirection: node_opcua_data_model_1.BrowseDirection.Inverse,
            includeSubtypes: true,
            nodeClassMask: 0,
            nodeId,
            referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasSubtype"),
            resultMask: 0xffff
        });
    }
    if (!browseResult.references || browseResult.references.length === 0) {
        return new node_opcua_types_1.ReferenceDescription({});
    }
    return browseResult.references[0];
}
async function getTypeDefinition(session, nodeId) {
    const browseResult = await session.browse({
        browseDirection: node_opcua_data_model_1.BrowseDirection.Forward,
        includeSubtypes: true,
        nodeClassMask: 0,
        nodeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasTypeDefinition"),
        resultMask: 0xffff
    });
    if (!browseResult.references || browseResult.references.length === 0) {
        debugLog("no subtype", nodeId.toString());
        throw new Error("No subtype");
    }
    return browseResult.references[0];
}
async function readBrowseName(session, nodeId) {
    const nodeToRead = {
        attributeId: node_opcua_data_model_1.AttributeIds.BrowseName,
        nodeId
    };
    const dataValue = await session.read(nodeToRead);
    // istanbul ignore next
    if (dataValue.statusCode.isNotGood()) {
        return new node_opcua_data_model_1.QualifiedName({ name: "", namespaceIndex: 0 });
    }
    const dataTypeName = dataValue.value.value;
    return dataTypeName;
}
// see also client-proxy
async function _convertNodeIdToDataTypeAsync(session, dataTypeId) {
    const dataTypeName = await readBrowseName(session, dataTypeId);
    let dataType;
    if (dataTypeId.namespace === 0 && dataTypeId.value === node_opcua_constants_1.DataTypeIds.Number) {
        return {
            type: "genericNumber",
            dataTypeCombination: "DataType.Float | DataType.Double | " +
                "DataType.UInt64 | DataType.UInt32 | DataType.UInt16 |  DataType.SByte | " +
                "DataType.Int64 | DataType.Int32 | DataType.Int16 | DataType.Byte"
        };
    }
    if (dataTypeId.namespace === 0 && dataTypeId.value === node_opcua_constants_1.DataTypeIds.Integer) {
        return {
            type: "genericNumber",
            dataTypeCombination: "DataType.Int64 | DataType.Int32 | DataType.Int16  | DataType.SByte"
        };
    }
    if (dataTypeId.namespace === 0 && dataTypeId.value === node_opcua_constants_1.DataTypeIds.UInteger) {
        return {
            type: "genericNumber",
            dataTypeCombination: "DataType.UInt64 | DataType.UInt32 | DataType.UInt16  | DataType.Byte"
        };
    }
    if (dataTypeId.namespace === 0 && dataTypeId.value === node_opcua_constants_1.DataTypeIds.Enumeration) {
        return { type: "enum", dataType: node_opcua_variant_1.DataType.Int32, enumerationId: dataTypeId };
    }
    if (dataTypeId.namespace === 0 && node_opcua_variant_1.DataType[dataTypeId.value]) {
        dataType = dataTypeId.value;
        return { type: "basic", dataType };
    }
    /// example => Duration (i=290) => Double (i=11)
    // read subTypeOf
    const nodeToBrowse = {
        browseDirection: node_opcua_data_model_1.BrowseDirection.Inverse,
        nodeId: dataTypeId,
        referenceTypeId: (0, node_opcua_nodeid_1.resolveNodeId)("HasSubtype"),
        resultMask: 0xff
    };
    // tslint:disable:no-shadowed-variable
    const browseResult = await session.browse(nodeToBrowse);
    const references = browseResult.references;
    if (!references || references.length !== 1) {
        throw new Error("cannot find SuperType of " + dataTypeName.toString());
    }
    const nodeId = references[0].nodeId;
    const info = await _convertNodeIdToDataTypeAsync(session, nodeId);
    if (info.type == "enum" && info.enumerationId) {
        return { ...info, enumerationId: dataTypeId, enumerationType: dataTypeName.name };
    }
    return info;
}
async function getChildrenOrFolderElements(session, nodeId) {
    const c1 = await getChildren(session, nodeId);
    const c2 = await getFolderElements(session, nodeId);
    return [].concat(c1, c2);
}
async function getValueRank(session, nodeId) {
    const valueRankDataValue = await session.read({ nodeId, attributeId: node_opcua_data_model_1.AttributeIds.ValueRank });
    return valueRankDataValue.value.value;
}
//# sourceMappingURL=utils.js.map